#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteIdentifier.h"
#include "KiteTree.h"
#include "KiteList.h"
#include "KiteBoolean.h"

using namespace KiteVM;

KiteReference<KiteObject> KiteCode_eval(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	vm.syms.removeContext();
	KiteReference<KiteObject> ret(dynamic_cast<KiteCode&>(*thisObj).call(vm, param));
	vm.syms.addContext();
	return ret;
}

KiteReference<KiteObject> KiteCode_catch(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	vm.syms.removeContext();
	KiteReference<KiteObject> ret(dynamic_cast<KiteCode&>(*thisObj).call(vm, param));
	vm.syms.addContext();
	
	if (typeid(*ret) == typeid(KiteTree)) {
		// Run stuff in catch block.
		vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("exception")), ret);
		return dynamic_cast<KiteCode&>(*ret).call(vm, param);
	}

	return ret;
}

KiteReference<KiteObject> KiteCode_while(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	vm.syms.removeContext();
	KiteReference<KiteObject> ret, cond;
	
begin:
	ret = dynamic_cast<KiteCode&>(*thisObj).call(vm, param);
	if (typeid(*ret) == typeid(KiteTree)) {
		// Exception; return now
		vm.syms.addContext();
		return ret;
	}

	// Loop conditional
	cond = dynamic_cast<KiteCode&>(*param).call(vm, new KiteList());
	if (typeid(*cond) != typeid(KiteBoolean)) {
		// Condition must return boolean
		vm.syms.addContext();
		return KiteReference<KiteObject>(new KiteTree("While condition must return boolean"));
	}

	// Continue if condition is true
	if (dynamic_cast<KiteBoolean&>(*cond).boolVal()) goto begin;

	// Else, return what the final iteration returned.
	vm.syms.addContext();
	return ret;
}

KiteReference<KiteObject> KiteCode_for(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteReference<KiteObject> ret, newlist, item;
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator i;
	
	// Convert parameter into a singleton list, if needed
	if (typeid(*param) != typeid(KiteList)) {
		newlist = KiteReference<KiteObject>(new KiteList(param));
	} else {
		newlist = param;
	}

	// Get items from list and iterate through them
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort > &items = dynamic_cast<KiteList&>(*newlist).getItems();
	for (i = items.begin(); i != items.end(); i++) {
		vm.syms.addContext();
		vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("key")), (*i).first);
		vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("value")), (*i).second);

		// Call function on iterated stuff
		ret = dynamic_cast<KiteCode&>(*thisObj).call(vm, new KiteList());
		if (typeid(*ret) == typeid(KiteTree)) {
			// Exception; return now
			vm.syms.removeContext();
			return ret;
		}
		vm.syms.removeContext();
	}

	// Return what the final iteration returned.
	return ret;
}

void KiteCode::initialize() {
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("eval"))] = KiteReference<KiteObject>(new KiteCode(KiteCode_eval));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("catch"))] = KiteReference<KiteObject>(new KiteCode(KiteCode_catch));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("while"))] = KiteReference<KiteObject>(new KiteCode(KiteCode_while));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("for"))] = KiteReference<KiteObject>(new KiteCode(KiteCode_for));
}

KiteReference<KiteObject> KiteCode::call(KiteVM::VM &vm, KiteReference<KiteObject> param) {
	if (funcPtr) {
		return (*funcPtr)(vm, param);
	} else {
		vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("args")), param);

		// Save # of items on stack so we can return the stack to how it was
		int items = vm.stack.size();

		KiteReference<KiteObject> ret = vm.run(instr);

		// Return stack to how it was before
		for (int i = vm.stack.size(); i > items; i--) {
			vm.stack.pop();
		}

		return ret;
	}
}

KiteCode::~KiteCode() {
	// nothing needed
}

//	virtual KiteReference<KiteObject> dereference(VM &vm, KiteIdentifier &id);
