#ifndef KITEFILE_H
#define KITEFILE_H

#include <string>
#include <iostream>
#include <fstream>
#include "KiteObject.h"
#include "KiteList.h"

using namespace std;

class KiteFile : public KiteObject {
public:
	KiteFile() : KiteObject(true), out(new fstream()), in(new fstream()) {
        inherits = KiteObject::getClassinfo();
        type = "file";
		initialize();
        mode = 3;
	}

	KiteFile(istream *tfs) : KiteObject(true), out(new fstream()), in(tfs), instance(true) { 
        type = "file";
		initialize();
        inherits = KiteObject::getClassinfo();
        mode = 1;
	}

	KiteFile(ostream *tfs) : KiteObject(true), out(tfs), in(new fstream()), instance(true) { 
        type = "file";
		initialize();
        inherits = KiteObject::getClassinfo();
        mode = 2;
	}

	KiteFile(VM &vm, KiteReference<KiteObject> param) 
		: KiteObject(true), out(new fstream()), in(new fstream()), instance(true) {
			inherits = KiteObject::getClassinfo();
            type = "file";
			initialize();
            mode = 3;

			KiteFile &fp = *this;

			if (typeid(*param) == typeid(KiteList)) {
				// Multiple parameters
				KiteList &list = dynamic_cast<KiteList&>(*param);
				KiteReference<KiteObject> filename(coerceString(vm, list[KiteReference<KiteObject>(new KiteNumber(0))]));
				KiteReference<KiteObject> mode(coerceString(vm, list[KiteReference<KiteObject>(new KiteNumber(1))]));
				std::string fm = dynamic_cast<KiteString&>(*mode).getString();
				ios::openmode m;

				if (fm[0] == 'r') m |= ios::in;
				else if (fm[0] == 'w') m |= ios::out;
				else if (fm[0] == 'a') m |= ios::app;

				if (fm[0] == 'r') {
					dynamic_cast<fstream&>(fp.getInput()).open(dynamic_cast<KiteString&>(*filename).getString().c_str(), m);
				} else {
					dynamic_cast<fstream&>(fp.getOutput()).open(dynamic_cast<KiteString&>(*filename).getString().c_str(), m);
				}
			} else {
				KiteReference<KiteObject> filename(coerceString(vm, param));
				dynamic_cast<fstream&>(fp.getInput()).open(dynamic_cast<KiteString&>(*filename).getString().c_str(), ios::in);
			}
	}

	virtual ~KiteFile() {
        if (mode & 1) delete out;
        if (mode & 2) delete in;
		// Seems to auto-close when the destructor's called. Good.
		/*if (dynamic_cast<fstream&>(in).is_open()) {
			dynamic_cast<fstream&>(in).close();
		}
		if (dynamic_cast<fstream&>(out).is_open()) {
			dynamic_cast<fstream&>(out).close();
		}*/
	}

	istream &getInput() { return *in; }
	ostream &getOutput() { return *out; }

	/*virtual KiteReference<KiteObject> call(VM &vm, KiteObject &param);
	virtual KiteReference<KiteObject> dereference(VM &vm, KiteIdentifier &id);*/
private:
	ostream *out;
	istream *in;
	bool instance;
    int mode;
	void initialize();
};

#endif // KITEFILE_H

