#include <sstream>
#include "KiteVM.h"
#include "KiteList.h"
#include "KiteString.h"
#include "KiteNumber.h"
#include "KiteReference.h"
#include "KiteCode.h"
#include "KiteBoolean.h"
#include "KiteTree.h"

using namespace KiteVM;
using namespace std;

KiteReference<KiteObject> KiteList_opplus(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &list1 = dynamic_cast<KiteList&>(*thisObj);
	KiteList *list2;

	if (typeid(*param) == typeid(KiteList)) {
		list2 = dynamic_cast<KiteList*>(param.getObj());
	} else {
		list2 = new KiteList(param);
	}
	
	KiteReference<KiteObject> ret(list1.merge(*list2));
	if (typeid(*param) != typeid(KiteList)) delete list2;
	return ret;
}

KiteReference<KiteObject> KiteList_intersect(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &list1 = dynamic_cast<KiteList&>(*thisObj);
	KiteList *list2;

	if (typeid(*param) == typeid(KiteList)) {
		list2 = dynamic_cast<KiteList*>(param.getObj());
	} else {
		list2 = new KiteList(param);
	}

	KiteList *ret = new KiteList();
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator i = list1.getItems().begin(), j;
	int idx = 0;
	for ( ; i != list1.getItems().end(); i++) {
		if (typeid(*(*i).first) == typeid(KiteNumber)) {
			j = list2->getItems().begin();
			bool insert = false;
			for ( ; j != list2->getItems().end(); j++) {
				if (valuesEqual(vm, (*j).second, (*i).second)) insert = true;
			}
			if (insert) ret->getItems()[KiteReference<KiteObject>(new KiteNumber(idx++))] = (*i).second;
		} else {
			if (list2->getItems().find((*i).first) != list2->getItems().end()) {
				if (valuesEqual(vm, (*i).second, (*list2->getItems().find((*i).first)).second)) {
					ret->getItems()[(*i).first] = (*i).second;
				}
			}
		}
	}

	KiteReference<KiteObject> ref(ret);
	if (typeid(*param) != typeid(KiteList)) delete list2;
	return ref;
}

KiteReference<KiteObject> KiteList_except(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &list1 = dynamic_cast<KiteList&>(*thisObj);
	KiteList *list2;

	if (typeid(*param) == typeid(KiteList)) {
		list2 = dynamic_cast<KiteList*>(param.getObj());
	} else {
		list2 = new KiteList(param);
	}

	KiteList *ret = new KiteList();
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator i = list1.getItems().begin(), j;
	int idx = 0;
	bool insert;
	for ( ; i != list1.getItems().end(); i++) {
		if (typeid(*(*i).first) == typeid(KiteNumber)) {
			j = list2->getItems().begin();
			insert = true;
			for ( ; j != list2->getItems().end(); j++) {
				if (valuesEqual(vm, (*j).second, (*i).second)) insert = false;
			}
			if (insert) ret->getItems()[KiteReference<KiteObject>(new KiteNumber(idx++))] = (*i).second;
		} else {
			insert = true;
			if (list2->getItems().find((*i).first) != list2->getItems().end()) {
				if (valuesEqual(vm, (*i).second, (*list2->getItems().find((*i).first)).second)) insert = false;
			}
			if (insert)	ret->getItems()[(*i).first] = (*i).second;
		}
	}

	KiteReference<KiteObject> ref(ret);
	if (typeid(*param) != typeid(KiteList)) delete list2;
	return ref;
}

KiteReference<KiteObject> KiteList_opderef(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteList &list = dynamic_cast<KiteList&>(*thisObj);

	return list[param];
}

KiteList *KiteList::merge(KiteList &rhs) {
	KiteList *newlist = new KiteList(*this);

	// First, find the largest numeric index in the first list
	int n = -1;
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort > items = getItems();
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator i = items.begin();
	for ( ; i != items.end(); i++) {
		if (typeid(*(*i).first) == typeid(KiteNumber)) {
			if (n < dynamic_cast<KiteNumber&>(*(*i).first).intValue()) {
				n = dynamic_cast<KiteNumber&>(*(*i).first).intValue();
			}
		}
		//newlist->getItems()[(*i).first] = (*i).second;
	}

	// Insert items one at a time now, incrementing by the largest index when needed.
	// Do not insert items with keys that already exist.
	i = rhs.getItems().begin();
	for ( ; i != rhs.getItems().end(); i++) {
		if (typeid(*(*i).first) == typeid(KiteNumber)) {
			KiteReference<KiteObject> newkey(new KiteNumber(dynamic_cast<KiteNumber&>(*(*i).first).intValue() + n + 1));
			newlist->items[newkey] = (*i).second;
		} else {
			newlist->items[(*i).first] = (*i).second;
		}
	}

	return newlist;
}

KiteReference<KiteObject> KiteList_string(VM &vm, KiteReference<KiteObject> param) {
	KiteList &lst = dynamic_cast<KiteList&>(*THIS_OBJECT());
	stringstream str;

	str << "[";
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator i = lst.getItems().begin();
	for ( ; i != lst.getItems().end(); i++) {
		KiteReference<KiteObject> k(coerceString(vm, (*i).first));
		KiteReference<KiteObject> v(coerceString(vm, (*i).second));
		KiteString &key = dynamic_cast<KiteString&>(*k);
		KiteString &val = dynamic_cast<KiteString&>(*v);
		str << key.getString() << ":" << val.getString() << " ";
	}
	str << "]";

	return KiteReference<KiteObject>(new KiteString(str.str()));
}

KiteReference<KiteObject> KiteList_print(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> t = coerceString(vm, THIS_OBJECT());
	KiteString &lst = dynamic_cast<KiteString&>(*t);
	cout << lst.getString() << endl;
	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteList_decide(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> to(THIS_OBJECT());
	KiteList &lst = dynamic_cast<KiteList&>(*to);
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort >::iterator i = lst.getItems().begin();
	stringstream err;

	// Go through each condition. If true, execute the code object next door.
	int condnum = 1;
	for ( ; i != lst.getItems().end(); i++) {
		if (typeid(*(*i).second) != typeid(KiteBoolean)) {
			// Not a boolean expression
			err << "Condition " << condnum << " must evaluate to boolean value";
			return KiteReference<KiteObject>(new KiteTree(err.str()));
		}

		KiteBoolean &b = dynamic_cast<KiteBoolean&>(*(*i).second);
		if (b.boolVal()) {
			// Condition evaluates to true, execute code block.
			i++;
			KiteReference<KiteObject> code((*i).second);
			if (typeid(*code) == typeid(KiteCode)) {
				vm.syms.removeContext();
				KiteReference<KiteObject> ret((*code).call(vm, param));
				vm.syms.addContext();
				return ret;
			} else {
				return (*i).second;
			}
		}

		i++; // Skip past and go on to the next condition
	}

	// Return false; none of the conditions passed.
	return KiteReference<KiteObject>(new KiteBoolean(false));
}

void KiteList::initialize() {
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op+"))] = KiteReference<KiteObject>(new KiteCode(KiteList_opplus));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("opor"))] = KiteReference<KiteObject>(new KiteCode(KiteList_opplus));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("opand"))] = KiteReference<KiteObject>(new KiteCode(KiteList_intersect));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op/"))] = KiteReference<KiteObject>(new KiteCode(KiteList_except));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("opderef"))] = KiteReference<KiteObject>(new KiteCode(KiteList_opderef));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("string"))] = KiteReference<KiteObject>(new KiteCode(KiteList_string));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("print"))] = KiteReference<KiteObject>(new KiteCode(KiteList_print));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("decide"))] = KiteReference<KiteObject>(new KiteCode(KiteList_decide));
}