#ifndef KITELIST_H
#define KITELIST_H

#include <map>
#include "KiteReference.h"
#include "KiteObject.h"
#include "KiteKeyValPair.h"
#include "KiteNumber.h"
#include "KiteIdentifier.h"

using namespace std;

struct list_sort {
	bool operator()(const KiteReference<KiteObject> &x1, const KiteReference<KiteObject> &x2) const {
		if (typeid(*x1) == typeid(KiteNumber) && typeid(*x2) != typeid(KiteNumber)) {
			// Numbers before everything else
			return true;
		} else if (typeid(*x1) != typeid(KiteNumber) && typeid(*x2) == typeid(KiteNumber)) {
			return false;
		} else if (typeid(*x1) == typeid(KiteNumber) && typeid(*x2) == typeid(KiteNumber)) {
			return dynamic_cast<KiteNumber&>(*x1).intValue() < dynamic_cast<KiteNumber&>(*x2).intValue();
		} else {
			return *x1 < *x2;
		}
	}
};

class KiteList : public KiteObject {
public:
	KiteList()
		: KiteObject(true) {
        inherits = KiteObject::getClassinfo();
        type = "list";
		initialize();
	}

	KiteList(const KiteList &obj) : KiteObject(true), instance(true), items(obj.items) {
        inherits = KiteObject::getClassinfo();
        type = "list";
		initialize();
    }

	KiteList(KiteReference<KiteObject> obj) : KiteObject(true), instance(true) {
        inherits = KiteObject::getClassinfo();
        type = "list";
		initialize();
		if (typeid(*obj) == typeid(KiteKeyValPair)) {
			KiteKeyValPair a = dynamic_cast<KiteKeyValPair&>(*obj);
			items[a.first] = a.second;
		} else {
			items[KiteReference<KiteObject>(new KiteNumber("0"))] = obj;
		}
	}

	/*virtual KiteReference<KiteObject> call(VM &vm, KiteObject &param);
	virtual KiteReference<KiteObject> dereference(VM &vm, KiteIdentifier &id);*/

	KiteList *merge(KiteList &rhs);

	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort > &getItems() { return items; }

	KiteReference<KiteObject> operator[](const KiteReference<KiteObject> &idx) {
		return items[idx];
	}

private:
	map<KiteReference<KiteObject>, KiteReference<KiteObject>, list_sort > items;
	bool instance;
	void initialize();
};

#endif // KITELIST_H
