#include <stdlib.h>
#include <sstream>
#include "KiteVM.h"
#include "KiteNumber.h"
#include "KiteString.h"
#include "KiteReference.h"
#include "KiteIdentifier.h"
#include "KiteCode.h"
#include "KiteBoolean.h"
#include "KiteTree.h"
#include "KiteList.h"

using namespace KiteVM;

KiteReference<KiteObject> KiteNumber_opplus(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	if (number1.isFloat() || number2.isFloat()) {
		return KiteReference<KiteObject>(new KiteNumber(number1.floatValue() + number2.floatValue()));
	} else {
		return KiteReference<KiteObject>(new KiteNumber(number1.intValue() + number2.intValue()));
	}
}

KiteReference<KiteObject> KiteNumber_opminus(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	if (number1.isFloat() || number2.isFloat()) {
		return KiteReference<KiteObject>(new KiteNumber(number1.floatValue() - number2.floatValue()));
	} else {
		return KiteReference<KiteObject>(new KiteNumber(number1.intValue() - number2.intValue()));
	}
}

KiteReference<KiteObject> KiteNumber_opmultiply(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	if (number1.isFloat() || number2.isFloat()) {
		return KiteReference<KiteObject>(new KiteNumber(number1.floatValue() * number2.floatValue()));
	} else {
		return KiteReference<KiteObject>(new KiteNumber(number1.intValue() * number2.intValue()));
	}
}

KiteReference<KiteObject> KiteNumber_opdivide(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	if ((number2.isFloat() && number2.floatValue() == 0) ||
		(!number2.isFloat() && number2.intValue() == 0)) {
		// Divide by 0 attempted. Bad dog!
		return KiteReference<KiteObject>(new KiteTree("Division by zero"));
	}

	return KiteReference<KiteObject>(new KiteNumber(number1.floatValue() / number2.floatValue()));
}

KiteReference<KiteObject> KiteNumber_opmod(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	if ((number2.isFloat() && number2.floatValue() == 0) ||
		(!number2.isFloat() && number2.intValue() == 0)) {
		// Divide by 0 attempted. Bad dog!
		return KiteReference<KiteObject>(new KiteTree("Division by zero"));
	}

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() % number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_opleftshift(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() << number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_oprightshift(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() >> number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_print(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));

	if (number.isFloat()) {
		cout << number.floatValue() << endl;
	} else {
		cout << number.intValue() << endl;
	}

	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteNumber_oplessthan(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(number1 < number2));
}

KiteReference<KiteObject> KiteNumber_oplessequals(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(number1 < number2 || number1 == number2));
}

KiteReference<KiteObject> KiteNumber_opequals(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(number1 == number2));
}

KiteReference<KiteObject> KiteNumber_opnotequals(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(!(number1 == number2)));
}

KiteReference<KiteObject> KiteNumber_opgreaterequals(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(!(number1 < number2)));
}

KiteReference<KiteObject> KiteNumber_opgreaterthan(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteBoolean(!(number1 < number2) && !(number1 == number2)));
}

KiteReference<KiteObject> KiteNumber_opand(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() & number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_opor(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() | number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_opxor(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	KiteNumber &number2 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, param));

	return KiteReference<KiteObject>(new KiteNumber(number1.intValue() ^ number2.intValue()));
}

KiteReference<KiteObject> KiteNumber_opnot(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));

	return KiteReference<KiteObject>(new KiteNumber(!number1.intValue()));
}

KiteReference<KiteObject> KiteNumber_number(VM &vm, KiteReference<KiteObject> param) {
	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteNumber_string(VM &vm, KiteReference<KiteObject> param) {
	KiteNumber &number1 = dynamic_cast<KiteNumber&>(*coerceNumber(vm, THIS_OBJECT()));
	stringstream str;
	string retstr;

	if (number1.isFloat()) str << number1.floatValue();
	else str << number1.intValue();

	str >> retstr;
	return KiteReference<KiteObject>(new KiteString(retstr));
}

KiteReference<KiteObject> KiteNumber_range(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> from(coerceNumber(vm, THIS_OBJECT()));
    KiteList *ret = new KiteList();
    KiteReference<KiteNumber> cur;
    
    KiteReference<KiteObject> to(coerceNumber(vm, (typeid(*param) == typeid(KiteList) ? dynamic_cast<KiteList&>(*param).getItems()[KiteReference<KiteObject>(new KiteNumber(0))] : param)));
    KiteReference<KiteObject> step;
    if (typeid(*param) == typeid(KiteList)) {
        step = dynamic_cast<KiteList&>(*param).getItems()[KiteReference<KiteObject>(new KiteNumber(1))];
    } else {
        step = KiteReference<KiteObject>(new KiteNumber(1));
    }
    
    while(1) {
        vm.syms.addContext();
        vm.syms.addVar(KiteReference<KiteIdentifier>(new KiteIdentifier("this")), from);
        KiteReference<KiteObject> comp((*((*from).dereference(vm, KiteReference<KiteIdentifier>(new KiteIdentifier("op<="))))).call(vm, to));
        if (!dynamic_cast<KiteBoolean&>(*comp).boolVal()) break;
        KiteList ele(from);
        KiteList *ret2 = ret->merge(ele);
        delete ret;
        ret = ret2;
        from = (*(*from).dereference(vm, KiteReference<KiteIdentifier>(new KiteIdentifier("op+")))).call(vm, step);
        vm.syms.removeContext();
    }
    
    return KiteReference<KiteObject>(ret);
}


KiteReference<KiteObject> KiteNumber_random(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> to(coerceNumber(vm, THIS_OBJECT()));
    static bool random_initialized = false;
    
    if (!random_initialized) {
        random_initialized = true;
        srandom(time(NULL));
    }
    
    return KiteReference<KiteObject>(new KiteNumber((int)(random() % dynamic_cast<KiteNumber&>(*to).intValue())));
}

void KiteNumber::initialize() {
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("number"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_number));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op+"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opplus));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op-"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opminus));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op*"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opmultiply));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op/"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opdivide));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("opmod"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opmod));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op<<"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opleftshift));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op>>"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_oprightshift));

	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op<"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_oplessthan));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op<="))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_oplessequals));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op=="))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opequals));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op!="))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opnotequals));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op>="))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opgreaterequals));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op>"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opgreaterthan));

	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("opand"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opand));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("opor"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opor));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("opnot"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opnot));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("xor"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_opxor));

	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("string"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_string));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("print"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_print));
    
    syms[KiteReference<KiteIdentifier>(new KiteIdentifier("range"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_range));
    syms[KiteReference<KiteIdentifier>(new KiteIdentifier("random"))] = KiteReference<KiteObject>(new KiteCode(KiteNumber_random));
}
