#ifndef KITEOBJECT_H
#define KITEOBJECT_H

#include <map>
#include "KiteReference.h"

using namespace std;

class KiteIdentifier;
class KiteList;

extern bool compareIdentifiers(const KiteReference<KiteIdentifier> &r1, const KiteReference<KiteIdentifier> &r2);

struct KiteIdentifierCompare {
    bool operator()(const KiteReference<KiteIdentifier> &r1, const KiteReference<KiteIdentifier> &r2) const {
		return compareIdentifiers(r1, r2);
    }
};

class KiteObject {
public:
	KiteObject(); // default, "class info" constructor
    KiteObject(bool init);
	KiteObject(KiteVM::VM &vm, KiteReference<KiteObject> param); // class|new
	virtual ~KiteObject();

	bool isInstance() { return instance; }

	// Variable lookup/execute
	virtual KiteReference<KiteObject> call(KiteVM::VM &vm, KiteReference<KiteObject> param);
	virtual KiteReference<KiteObject> dereference(KiteVM::VM &vm, KiteReference<KiteIdentifier> id);

	virtual bool operator<(const KiteObject &rhs) const {
		return this < &rhs;
	}

	virtual bool operator==(const KiteObject &rhs) const {
		return this == &rhs;
	}

	static KiteReference<KiteObject> &getClassinfo() {
        if (classinfo.getObj() == NULL) {
            classinfo = new KiteObject();
        }

		return classinfo;
	}

	virtual KiteReference<KiteObject>
	KiteObject::create(KiteVM::VM &vm, KiteReference<KiteObject> base, KiteReference<KiteObject> param);
	
	map<KiteReference<KiteIdentifier>, KiteReference<KiteObject>, KiteIdentifierCompare > getSyms() { return syms; }
	virtual KiteReference<KiteObject> getInherits() { return inherits; }

    string type;
protected:
	KiteReference<KiteObject> inherits; // object inherited from
	static KiteReference<KiteObject> classinfo;
	bool instance;
	static bool initialized;

	map<KiteReference<KiteIdentifier>, KiteReference<KiteObject>, KiteIdentifierCompare > syms;

	void initialize();
};

#endif // KITEOBJECT_H
