#ifndef KITEREFERENCE_H
#define KITEREFERENCE_H

#include <map>

class KiteIdentifier;

using namespace std;

template<class T>
class KiteReference {
public:
	KiteReference() : refs(new int(1)), object(new T()) {
		reflist.insert(pair<T* const, KiteReference<T>* >(object, this));
	}

	/*KiteReference(T &val)
		: refs(new int(1)),
		object(&val) {};*/

	KiteReference(T *val) {
        if (val) {
            refs = new int(1);
            object = val;
		    reflist.insert(pair<T* const, KiteReference<T>* >(object, this));
        } else {
            object = NULL;
            refs = NULL;
        }
    }

	KiteReference(const KiteReference<T> &obj)
		: refs(obj.refs),
		object(obj.object) { 
			if (obj.refs) {
				(*refs)++;
		        reflist.insert(pair<T* const, KiteReference<T>* >(object, this));
			}
	}

	virtual ~KiteReference() {
		destroy();
	};

	T *operator->() const {
		return object;
	};

	T &operator*() const {
		return *object;
	};

	KiteReference<T> &operator=(const KiteReference<T> &in) {
		destroy();
		refs = in.refs;
		object = in.object;
		reflist.insert(pair<T* const, KiteReference<T>* >(object, this));
		(*refs)++;
		return *this;
	}

	void replace(const KiteReference<T> &rhs) {
		destroy();
		typename multimap<T*, KiteReference<T>* >::iterator i = reflist.find(object);
		while (i != reflist.end()) {
			if ((*i).first != object) {
				break;
			}

			(*i).second->object = rhs.object;
			reflist.erase(i);
		    reflist.insert(pair<T* const, KiteReference<T>* >(rhs.object, (*i).second));
			i++;
		}

		refs = rhs.refs;
		(*refs)++;
	}

    KiteReference<KiteIdentifier> referenceIdentifier() {
        KiteReference<KiteIdentifier> ret(NULL);
        ret.refs = refs;
        ret.object = dynamic_cast<KiteIdentifier*>(object);
        (*ret.refs)++;

		//KiteReference<KiteIdentifier>::getReflist().insert(pair<KiteIdentifier* const, KiteReference<KiteIdentifier>* >(ret.getObj(), ret));
        return ret;
    }

	T *operator=(const T *in) {
		destroy();
		refs = new int(1);
		object = const_cast<T*>(in);

		reflist.insert(pair<T* const, KiteReference<T>* >(object, this));
		return object;
	}
    
	bool operator==(const KiteReference<T> &ptr) const {
		return *object == *ptr.object;
	}

	bool operator==(const T *ptr) const {
		return object == ptr;
	}

	bool operator<(const KiteReference<T> &ptr) const {
		return *object < *(ptr.object);
	}

	int *getRefs() const { return refs; }
	T *getObj() const { return object; }

	static multimap<T*, KiteReference<T>* > &getReflist() { return reflist; }
public:
	int *refs;
	T *object;

private:
	void destroy() {
        if (refs && object && reflist.count(object)) {
		    //(*refs) = reflist.count(object) - 1;
            (*refs)--;
            
			typename multimap<T*, KiteReference<T>* >::iterator i = reflist.find(object);
			while (i != reflist.end()) {
				if ((*i).first == object && (*i).second == this) {
					typename multimap<T*, KiteReference<T>* >::iterator j = i;
					i++;
					reflist.erase(j);
					continue;
				}
				i++;
			}

		    if (*refs == 0) {
			    delete refs;
			    delete object;
		    }
	    }
    }

	static multimap<T*, KiteReference<T>* > reflist;
};

template<class T>
multimap<T*, KiteReference<T>* > KiteReference<T>::reflist;

#endif // KITEREFERENCE_H
