#include <sstream>
#include "KiteLexer.hpp"
#include "KiteParser.hpp"
#include "KiteWalker.hpp"
#include "KiteVM.h"
#include "KiteNumber.h"
#include "KiteReference.h"
#include "KiteIdentifier.h"
#include "KiteCode.h"
#include "KiteBoolean.h"
#include "KiteTree.h"
#include "KiteString.h"
#include "KiteList.h"

using namespace KiteVM;
using namespace antlr;

KiteReference<KiteObject> KiteString_opplus(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string1 = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
	KiteReference<KiteObject> str2(coerceString(vm, param));
	KiteString &string2 = dynamic_cast<KiteString&>(*str2);

	return KiteReference<KiteObject>(new KiteString(string1.getString() + string2.getString()));
}

KiteReference<KiteObject> KiteString_opmultiply(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
	KiteReference<KiteObject> p2(coerceNumber(vm, param));
	KiteNumber &numtimes = dynamic_cast<KiteNumber&>(*p2);
	std::string ret;

	for (int i = 0; i < numtimes.intValue(); i++) {
		ret += string.getString();
	}

	return KiteReference<KiteObject>(new KiteString(ret));
}

KiteReference<KiteObject> KiteString_print(VM &vm, KiteReference<KiteObject> param) {
	KiteString &str = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));

	cout << str.getString() << endl;
	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteString_oplessthan(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string1 = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
	KiteReference<KiteObject> str2(coerceString(vm, param));
	KiteString &string2 = dynamic_cast<KiteString&>(*str2);

	return KiteReference<KiteObject>(new KiteBoolean(string1 < string2));
}

KiteReference<KiteObject> KiteString_oplessequals(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string1 = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
	KiteReference<KiteObject> str2(coerceString(vm, param));
	KiteString &string2 = dynamic_cast<KiteString&>(*str2);

	return KiteReference<KiteObject>(new KiteBoolean(string1 < string2 || string1 == string2));
}

KiteReference<KiteObject> KiteString_opequals(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string1 = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
	KiteReference<KiteObject> str2(coerceString(vm, param));
	KiteString &string2 = dynamic_cast<KiteString&>(*str2);

	return KiteReference<KiteObject>(new KiteBoolean(string1 == string2));
}

KiteReference<KiteObject> KiteString_opnotequals(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string1 = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
	KiteReference<KiteObject> str2(coerceString(vm, param));
	KiteString &string2 = dynamic_cast<KiteString&>(*str2);

	return KiteReference<KiteObject>(new KiteBoolean(!(string1 == string2)));
}

KiteReference<KiteObject> KiteString_opgreaterequals(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string1 = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
	KiteReference<KiteObject> str2(coerceString(vm, param));
	KiteString &string2 = dynamic_cast<KiteString&>(*str2);

	return KiteReference<KiteObject>(new KiteBoolean(!(string1 < string2)));
}

KiteReference<KiteObject> KiteString_opgreaterthan(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string1 = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
	KiteReference<KiteObject> str2(coerceString(vm, param));
	KiteString &string2 = dynamic_cast<KiteString&>(*str2);

	return KiteReference<KiteObject>(new KiteBoolean(!(string1 < string2) && !(string1 == string2)));
}

KiteReference<KiteObject> KiteString_number(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));

	if (string.getString().find(".") != std::string::npos) {
		return KiteReference<KiteObject>(new KiteNumber(atof(string.getString().c_str())));
	} else {
		return KiteReference<KiteObject>(new KiteNumber(atoi(string.getString().c_str())));
	}
}

KiteReference<KiteObject> KiteString_string(VM &vm, KiteReference<KiteObject> param) {
	return THIS_OBJECT();
}

KiteReference<KiteObject> KiteString_compile(VM &vm, KiteReference<KiteObject> param) {
	KiteString &string = dynamic_cast<KiteString&>(*coerceString(vm, THIS_OBJECT()));
    stringstream str;
    
    str.str(string.getString());
    
    KiteLexer lexer(str);
    KiteParser parser(lexer);
    ASTFactory my_factory;
	parser.initializeASTFactory( my_factory );
    parser.setASTFactory( &my_factory );
    parser.expr();
    RefAST t = parser.getAST();

	if (t.get()) {
		KiteWalker walker;
        Instruction *i = walker.expr(t);
        KiteReference<KiteObject> ret(new KiteCode(i));
        return ret;
    }
    
    return KiteReference<KiteObject>(new KiteTree("Parse error while compiling code"));
}

KiteReference<KiteObject> KiteString_eval(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> fxn = KiteString_compile(vm, param);
    
    // If there was a problem compiling, just return the exception.
    if (typeid(fxn) == typeid(KiteTree)) {
        return fxn;
    }
    
    // Otherwise, evaluate it.
    vm.syms.removeContext();
	KiteReference<KiteObject> ret(dynamic_cast<KiteCode&>(*fxn).call(vm, param));
	vm.syms.addContext();
	return ret;
}

KiteReference<KiteObject> KiteString_opderef(VM &vm, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> thisObj(THIS_OBJECT());
	KiteString &list = dynamic_cast<KiteString&>(*thisObj);
	KiteReference<KiteObject> p2(coerceNumber(vm, param));
	KiteNumber &idx = dynamic_cast<KiteNumber&>(*p2);
	char buf[2];

	buf[0] = list.getString()[idx.intValue() + 1];
	buf[1] = '\0';

	return KiteReference<KiteObject>(new KiteString(string(buf)));
}

KiteReference<KiteObject> KiteString_find(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());
    KiteReference<KiteObject> toFind(coerceString(vm, typeid(*param) == typeid(KiteList) ? dynamic_cast<KiteList&>(*param).getItems()[KiteReference<KiteObject>(new KiteNumber(0))] : param));
    KiteReference<KiteObject> start;
    if (typeid(*param) == typeid(KiteList)) {
        start = dynamic_cast<KiteList&>(*param).getItems()[KiteReference<KiteObject>(new KiteNumber(1))];
    } else {
        start = KiteReference<KiteObject>(new KiteNumber(0));
    }

    KiteString &str = dynamic_cast<KiteString&>(*toFind);
    int startPos = dynamic_cast<KiteNumber&>(*start).intValue();
    int loc = dynamic_cast<KiteString&>(*string).getString().find(str.getString(), startPos);
    
    return KiteReference<KiteObject>(new KiteNumber(loc));
}

KiteReference<KiteObject> KiteString_substring(VM &vm, KiteReference<KiteObject> param) {
    KiteReference<KiteObject> string(THIS_OBJECT());
    KiteReference<KiteObject> begin(coerceNumber(vm, typeid(*param) == typeid(KiteList) ? dynamic_cast<KiteList&>(*param).getItems()[KiteReference<KiteObject>(new KiteNumber(0))] : param));
    KiteReference<KiteObject> end;
    
    if (typeid(*param) == typeid(KiteList)) {
        end = dynamic_cast<KiteList&>(*param).getItems()[KiteReference<KiteObject>(new KiteNumber(1))];
    } else {
        end = KiteReference<KiteObject>(new KiteNumber(int(dynamic_cast<KiteString&>(*string).getString().size() - 1)));
    }

    int startPos = dynamic_cast<KiteNumber&>(*begin).intValue();
    int endPos = dynamic_cast<KiteNumber&>(*end).intValue();
    
    std::string res = dynamic_cast<KiteString&>(*string).getString().substr(startPos, endPos);
    
    return KiteReference<KiteObject>(new KiteString(res));
}

void KiteString::initialize() {
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("number"))] = KiteReference<KiteObject>(new KiteCode(KiteString_number));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op+"))] = KiteReference<KiteObject>(new KiteCode(KiteString_opplus));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op*"))] = KiteReference<KiteObject>(new KiteCode(KiteString_opmultiply));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("opderef"))] = KiteReference<KiteObject>(new KiteCode(KiteString_opderef));

	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op<"))] = KiteReference<KiteObject>(new KiteCode(KiteString_oplessthan));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op<="))] = KiteReference<KiteObject>(new KiteCode(KiteString_oplessequals));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op=="))] = KiteReference<KiteObject>(new KiteCode(KiteString_opequals));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op!="))] = KiteReference<KiteObject>(new KiteCode(KiteString_opnotequals));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op>="))] = KiteReference<KiteObject>(new KiteCode(KiteString_opgreaterequals));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("op>"))] = KiteReference<KiteObject>(new KiteCode(KiteString_opgreaterthan));

	//syms[KiteReference<KiteIdentifier>(new KiteIdentifier("xor"))] = KiteReference<KiteObject>(new KiteCode(KiteString_opxor));

	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("string"))] = KiteReference<KiteObject>(new KiteCode(KiteString_string));
	syms[KiteReference<KiteIdentifier>(new KiteIdentifier("print"))] = KiteReference<KiteObject>(new KiteCode(KiteString_print));
    
    syms[KiteReference<KiteIdentifier>(new KiteIdentifier("compile"))] = KiteReference<KiteObject>(new KiteCode(KiteString_compile));
    syms[KiteReference<KiteIdentifier>(new KiteIdentifier("eval"))] = KiteReference<KiteObject>(new KiteCode(KiteString_eval));
    
    syms[KiteReference<KiteIdentifier>(new KiteIdentifier("find"))] = KiteReference<KiteObject>(new KiteCode(KiteString_find));
    syms[KiteReference<KiteIdentifier>(new KiteIdentifier("substring"))] = KiteReference<KiteObject>(new KiteCode(KiteString_substring));
}
