#include <iostream>
#include <fstream>
#include "KiteLexer.hpp"
#include "KiteParser.hpp"
#include "KiteWalker.hpp"
#include "KiteIdentifier.h"
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteTree.h"

using namespace std;
using namespace antlr;

int main(int argc, char **argv) {
	try {
        ASTFactory my_factory;
        RefAST t;
        
        // Use standard input if nothing passed in
        ifstream input;
        if (argc == 2) {
            input.open(argv[1], ios::in);
        	KiteLexer lexer(input);
            KiteParser parser(lexer);
			parser.initializeASTFactory( my_factory );
            parser.setASTFactory( &my_factory );
        	parser.expr();
            input.close();
       		t = parser.getAST();
        } else {
            KiteLexer lexer(cin);
            KiteParser parser(lexer);
			parser.initializeASTFactory( my_factory );
            parser.setASTFactory( &my_factory );
        	parser.expr();
            t = parser.getAST();
        }
        
		if (t.get()) {
			//cout << t->toStringTree() << endl;
			KiteWalker walker;
			Instruction *i = walker.expr(t);
			//i->outTree();
			KiteVM::VM vm;
			KiteReference<KiteObject> ret = vm.run(i);
            
            if (typeid(*ret) == typeid(KiteTree)) {
                KiteTree &exc = dynamic_cast<KiteTree&>(*ret);
                cerr << "Exception in code: " << exc.getMessage() << endl;
            }
            
			i->deallocate();
		}

	} catch(exception &e) {
		cerr << "exception: " << e.what() << endl;
	}

    return 0;
}
